"""Helper utilities for RDF converters."""

from .conversion_utils import (
    build_uri,
    ensure_uri,
    format_error,
    map_literal,
    report_progress,
)

__all__ = [
    "build_uri",
    "ensure_uri",
    "format_error",
    "map_literal",
    "report_progress",
]



"""RDF graph serialization helper."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from rdflib import Graph

from libraries.ontology.bnop.rdf_jena.constants.serialization_formats import (
    SUPPORTED_FORMATS,
    detect_format_from_filename,
)
from libraries.ontology.bnop.rdf_jena.object_models.rdf_graphs import RdfGraph

_FORMAT_TO_RDFLIB = {
    "turtle": "turtle",
    "rdf/xml": "pretty-xml",
    "json-ld": "json-ld",
    "n-triples": "nt",
}


@dataclass
class Serializer:
    """Serialize :class:`RdfGraph` to various formats."""

    def _build_graph(self, graph: RdfGraph) -> Graph:
        rdflib_graph = Graph()
        for prefix, uri in graph.namespaces.items():
            rdflib_graph.bind(prefix, uri)
        data = graph.to_n_triples()
        if data:
            rdflib_graph.parse(data=data, format="nt")
        return rdflib_graph

    def to_string(self, graph: RdfGraph, format_name: str) -> str:
        """Return ``graph`` serialized to ``format_name``.

        Raises
        ------
        ValueError
            If ``format_name`` is unsupported.

        """
        if format_name not in SUPPORTED_FORMATS:
            raise ValueError("unsupported format")
        rdflib_graph = self._build_graph(graph)
        rdflib_format = _FORMAT_TO_RDFLIB[format_name]
        return rdflib_graph.serialize(format=rdflib_format)

    def to_file(
        self,
        graph: RdfGraph,
        path: str | Path,
        *,
        format_name: str | None = None,
    ) -> None:
        """Write ``graph`` to ``path`` using ``format_name``.

        If ``format_name`` is ``None`` it is inferred from ``path``.
        """
        fmt = format_name or detect_format_from_filename(str(path))
        if fmt is None:
            raise ValueError("format could not be inferred")
        data = self.to_string(graph, fmt)
        Path(path).write_text(data, encoding="utf-8")



"""Object models used across RDF-Jena components."""

from .jena_connections import (
    JenaConnectionError,
    JenaConnections,
)
from .jena_databases import JenaDatabases
from .jena_sessions import (
    JenaSessions,
    TransactionError,
)
from .query_results import QueryResults
from .rdf_graphs import RdfGraph
from .triples import Triple

__all__ = [
    "JenaConnectionError",
    "JenaConnections",
    "JenaDatabases",
    "JenaSessions",
    "QueryResults",
    "RdfGraph",
    "TransactionError",
    "Triple",
]



from __future__ import annotations

from typing import Any

from libraries.ontology.bnop.rdf_jena.constants.standard_namespaces import (
    OWL_NS,
    RDF_NS,
    RDFS_NS,
)
from libraries.ontology.bnop.rdf_jena.converters.base_rdf_converter import (
    BaseRdfConverter,
)
from libraries.ontology.bnop.rdf_jena.converters.helpers.conversion_utils import (
    build_uri,
    map_literal,
)
from libraries.ontology.bnop.rdf_jena.object_models.rdf_graphs import RdfGraph
from libraries.ontology.bnop.rdf_jena.object_models.triples import Triple

try:  # pragma: no cover
    from nf_common_base.b_source.services.identification_services.b_identity_ecosystem.objects.bie_ids import (
        BieIds,
    )

    from libraries.interop_services.bclearer_interop_services.b_dictionary_service.objects.row_b_dictionaries import (
        RowBDictionaries,
    )
    from libraries.interop_services.bclearer_interop_services.b_dictionary_service.objects.table_b_dictionaries import (
        TableBDictionaries,
    )
except Exception:  # pragma: no cover
    TableBDictionaries = Any  # type: ignore[misc]
    RowBDictionaries = Any  # type: ignore[misc]

    class BieIds:  # type: ignore[too-few-public-methods]
        def __init__(self, value: object = "") -> None:
            self.value = str(value)

        def __str__(self) -> str:
            return self.value


class DatabaseToRdfConverter(BaseRdfConverter):
    """Convert database table schemas to RDF classes."""

    def __init__(self, namespace: str) -> None:
        super().__init__()
        if not namespace:
            raise ValueError("namespace required")
        self._namespace = namespace

    def _convert_item(self, item: TableBDictionaries, graph: RdfGraph) -> None:
        """Add schema triples for *item* to *graph*."""
        self._map_table_schema(item, graph)
        self._map_table_rows(item, graph)

    def _map_table_schema(self, table: TableBDictionaries, graph: RdfGraph) -> None:
        """Map *table* definition to an RDF class."""
        class_uri = build_uri(self._namespace, table.table_name)
        graph.add_triple(
            Triple(class_uri, build_uri(RDF_NS, "type"), build_uri(OWL_NS, "Class")),
        )

        rows = getattr(table, "dictionary", {})
        if not rows:
            return
        first_row = next(iter(rows.values()))
        columns = getattr(first_row, "dictionary", {})
        for column in columns:
            predicate = build_uri(self._namespace, column)
            graph.add_triple(
                Triple(
                    predicate,
                    build_uri(RDF_NS, "type"),
                    build_uri(RDF_NS, "Property"),
                ),
            )
            graph.add_triple(
                Triple(predicate, build_uri(RDFS_NS, "domain"), class_uri),
            )

    def _map_table_rows(
        self,
        table: TableBDictionaries,
        graph: RdfGraph,
    ) -> None:
        """Map *table* rows to RDF instances."""
        class_uri = build_uri(self._namespace, table.table_name)
        rows: dict[object, RowBDictionaries] = getattr(
            table,
            "dictionary",
            {},
        )
        for row_id, row in rows.items():
            subject = build_uri(self._namespace, str(row_id))
            graph.add_triple(
                Triple(
                    subject,
                    build_uri(RDF_NS, "type"),
                    class_uri,
                ),
            )
            columns = getattr(row, "dictionary", {})
            for column, value in columns.items():
                if value is None:
                    continue
                predicate = build_uri(self._namespace, column)
                if isinstance(value, BieIds):
                    obj = build_uri(self._namespace, str(value))
                else:
                    obj = map_literal(value)
                graph.add_triple(Triple(subject, predicate, obj))



from __future__ import annotations

from libraries.ontology.bnop.rdf_jena.constants.standard_namespaces import (
    RDF_NS,
)
from libraries.ontology.bnop.rdf_jena.object_models.triples import Triple

SHACL_NS = "http://www.w3.org/ns/shacl#"

NODE_SHAPE = f"{SHACL_NS}NodeShape"
PROPERTY_SHAPE = f"{SHACL_NS}PropertyShape"
PROPERTY = f"{SHACL_NS}property"
PATH = f"{SHACL_NS}path"
CLASS = f"{SHACL_NS}class"
MIN_COUNT = f"{SHACL_NS}minCount"
MAX_COUNT = f"{SHACL_NS}maxCount"


def build_node_shape(uri: str) -> Triple:
    """Return a triple declaring ``uri`` as a SHACL node shape."""
    return Triple(uri, f"{RDF_NS}type", NODE_SHAPE)


def link_property_shape(node_shape: str, prop_shape: str) -> Triple:
    """Return a triple linking ``prop_shape`` to ``node_shape``."""
    return Triple(node_shape, PROPERTY, prop_shape)


def build_property_shape(
    uri: str,
    path: str,
    *,
    class_uri: str | None = None,
    min_count: int | None = None,
    max_count: int | None = None,
) -> list[Triple]:
    """Return triples defining a SHACL property shape."""
    triples = [
        Triple(uri, f"{RDF_NS}type", PROPERTY_SHAPE),
        Triple(uri, PATH, path),
    ]
    if class_uri is not None:
        triples.append(Triple(uri, CLASS, class_uri))
    if min_count is not None:
        triples.append(Triple(uri, MIN_COUNT, str(min_count)))
    if max_count is not None:
        triples.append(Triple(uri, MAX_COUNT, str(max_count)))
    return triples



from dataclasses import dataclass


@dataclass
class Triple:
    """Representation of an RDF triple."""

    subject: str
    predicate: str
    object: str

    def __post_init__(self) -> None:
        self.validate()

    def validate(self) -> None:
        if not self.subject:
            raise ValueError("subject required")
        if not self.predicate:
            raise ValueError("predicate required")
        if not self.object:
            raise ValueError("object required")

    def to_n_triple(self) -> str:
        self.validate()
        return f"{self.subject} {self.predicate} {self.object} ."



"""OWL vocabulary constants and axiom builders."""

from __future__ import annotations

from libraries.ontology.bnop.rdf_jena.constants.standard_namespaces import (
    OWL_NS,
    RDF_NS,
    RDFS_NS,
)
from libraries.ontology.bnop.rdf_jena.object_models.triples import Triple

# OWL entity declarations
CLASS = f"{OWL_NS}Class"
OBJECT_PROPERTY = f"{OWL_NS}ObjectProperty"
DATATYPE_PROPERTY = f"{OWL_NS}DatatypeProperty"
ANNOTATION_PROPERTY = f"{OWL_NS}AnnotationProperty"

# Common OWL axioms
EQUIVALENT_CLASS = f"{OWL_NS}equivalentClass"
SAME_AS = f"{OWL_NS}sameAs"
SUBCLASS_OF = f"{RDFS_NS}subClassOf"
RDF_TYPE = f"{RDF_NS}type"


def declare_class(uri: str) -> Triple:
    """Return a triple declaring ``uri`` as an OWL class."""
    return Triple(uri, RDF_TYPE, CLASS)


def declare_object_property(uri: str) -> Triple:
    """Return a triple declaring ``uri`` as an OWL object property."""
    return Triple(uri, RDF_TYPE, OBJECT_PROPERTY)


def declare_datatype_property(uri: str) -> Triple:
    """Return a triple declaring ``uri`` as an OWL datatype property."""
    return Triple(uri, RDF_TYPE, DATATYPE_PROPERTY)


def build_subclass_axiom(subclass: str, superclass: str) -> Triple:
    """Return a triple stating that ``subclass`` is a subclass of ``superclass``."""
    return Triple(subclass, SUBCLASS_OF, superclass)


def build_equivalent_class_axiom(first: str, second: str) -> Triple:
    """Return a triple stating that ``first`` is equivalent to ``second``."""
    return Triple(first, EQUIVALENT_CLASS, second)


def build_same_as_axiom(first: str, second: str) -> Triple:
    """Return a triple stating that ``first`` and ``second`` denote the same individual."""
    return Triple(first, SAME_AS, second)



"""Expose RDF-Jena constant modules."""

from . import (
    rdf_data_types,
    serialization_formats,
    sparql_query_forms,
    standard_namespaces,
)

__all__ = [
    "rdf_data_types",
    "serialization_formats",
    "sparql_query_forms",
    "standard_namespaces",
]



"""RDF data type constants and utilities."""

import re
from datetime import date, datetime

RDF_NS = "http://www.w3.org/1999/02/22-rdf-syntax-ns#"
RDFS_NS = "http://www.w3.org/2000/01/rdf-schema#"
OWL_NS = "http://www.w3.org/2002/07/owl#"
XSD_NS = "http://www.w3.org/2001/XMLSchema#"

URI_PATTERN = re.compile(
    r"^[a-zA-Z][a-zA-Z0-9+.-]*://"
    r"[^\s<>\"{}|^`\\]*$",
)

RDF_LITERAL_URIS: dict[str, str] = {
    "string": f"{XSD_NS}string",
    "boolean": f"{XSD_NS}boolean",
    "integer": f"{XSD_NS}integer",
    "float": f"{XSD_NS}double",
    "date": f"{XSD_NS}date",
    "dateTime": f"{XSD_NS}dateTime",
}

PYTHON_TYPE_MAP: dict[type, str] = {
    str: "string",
    bool: "boolean",
    int: "integer",
    float: "float",
    datetime: "dateTime",
    date: "date",
}

ValueType = str | bool | int | float | date | datetime


def is_valid_uri(uri: str) -> bool:
    return bool(URI_PATTERN.match(uri))


def is_valid_literal(
    value: ValueType,
    literal: str,
) -> bool:
    expected = {
        "string": str,
        "boolean": bool,
        "integer": int,
        "float": float,
        "date": date,
        "dateTime": datetime,
    }[literal]
    if expected is int and isinstance(value, bool):
        return False
    return isinstance(value, expected)


def get_literal_uri(value: ValueType) -> str:
    for py, name in PYTHON_TYPE_MAP.items():
        if isinstance(value, py):
            if py is int and isinstance(value, bool):
                continue
            return RDF_LITERAL_URIS[name]
    raise TypeError("unsupported literal")



from __future__ import annotations

from collections.abc import Mapping
from csv import DictReader
from pathlib import Path

from libraries.ontology.bnop.rdf_jena.constants.rdf_data_types import (
    ValueType,
)
from libraries.ontology.bnop.rdf_jena.converters.base_rdf_converter import (
    BaseRdfConverter,
)
from libraries.ontology.bnop.rdf_jena.converters.helpers.conversion_utils import (
    build_uri,
    map_literal,
)
from libraries.ontology.bnop.rdf_jena.object_models.rdf_graphs import (
    RdfGraph,
)
from libraries.ontology.bnop.rdf_jena.object_models.triples import (
    Triple,
)


class CsvToRdfConverter(BaseRdfConverter):
    """Converter turning CSV rows into RDF."""

    def __init__(self, namespace: str) -> None:
        super().__init__()
        if not namespace:
            raise ValueError("namespace required")
        self._namespace = namespace

    def convert_file(self, path: str | Path) -> RdfGraph:
        """Convert CSV file at *path* to RDF."""
        with open(path, newline="", encoding="utf-8") as handle:
            reader = DictReader(handle)
            return self.convert(reader)

    def _convert_item(
        self,
        item: Mapping[str, ValueType],
        graph: RdfGraph,
    ) -> None:
        """Add triples for CSV *item* to *graph*."""
        if not isinstance(item, Mapping):
            raise TypeError("item must be mapping")
        identifier = str(item.get("id", self.processed))
        subject = build_uri(self._namespace, identifier)
        for column, value in item.items():
            if column == "id" or value is None:
                continue
            predicate = build_uri(self._namespace, column)
            obj = map_literal(value)
            graph.add_triple(Triple(subject, predicate, obj))



from dataclasses import dataclass

from bclearer_core.configurations.b_clearer_configurations.b_clearer_configurations import (
    BClearerConfigurations,
)


@dataclass
class JenaLoaderConfigurations(BClearerConfigurations):
    """Settings for bulk loading to Jena."""

    batch_size: int = 1000
    parallel_workers: int = 4
    streaming_threshold_mb: int = 100
    cache_size_mb: int = 500
    use_bulk_loader: bool = True

    def __post_init__(self) -> None:
        self.validate()

    def validate(self) -> None:
        if self.batch_size <= 0:
            raise ValueError("batch_size must be positive")
        if self.parallel_workers <= 0:
            raise ValueError(
                "parallel_workers must be positive",
            )
        if self.streaming_threshold_mb < 0:
            raise ValueError(
                "streaming_threshold_mb must be non-negative",
            )
        if self.cache_size_mb <= 0:
            raise ValueError("cache_size_mb must be positive")



"""Resource modules for RDF-Jena."""

from . import owl_vocabularies, prefix_mappings, query_templates, shacl_shapes

__all__ = [
    "owl_vocabularies",
    "prefix_mappings",
    "query_templates",
    "shacl_shapes",
]



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable

from libraries.ontology.bnop.rdf_jena.object_models.rdf_graphs import (
    RdfGraph,
)


class BaseRdfConverter(ABC):
    """Abstract converter producing RDF graphs."""

    def __init__(self) -> None:
        self._processed = 0

    @property
    def processed(self) -> int:
        """Number of items converted."""
        return self._processed

    def convert(self, source: Iterable[object]) -> RdfGraph:
        """Convert items from *source* to RDF."""
        if source is None:
            raise ValueError("source cannot be None")

        graph = RdfGraph()
        for item in source:
            try:
                self._convert_item(item, graph)
            except Exception as exc:  # pragma: no cover
                msg = f"Failed at item {self._processed}: {exc}"
                raise RuntimeError(msg) from exc
            self._processed += 1
        return graph

    @abstractmethod
    def _convert_item(self, item: object, graph: RdfGraph) -> None:
        """Add triples for *item* to *graph*."""



from __future__ import annotations

"""Service facade for RDF-Jena operations."""

from collections.abc import Iterable
from typing import Self

from libraries.ontology.bnop.rdf_jena.converters.base_rdf_converter import (
    BaseRdfConverter,
)
from libraries.ontology.bnop.rdf_jena.object_models.jena_databases import (
    JenaDatabases,
)
from libraries.ontology.bnop.rdf_jena.object_models.jena_sessions import (
    JenaSessions,
)
from libraries.ontology.bnop.rdf_jena.object_models.query_results import (
    QueryResults,
)
from libraries.ontology.bnop.rdf_jena.orchestrators.jena_data_load_orchestrator import (
    JenaDataLoadOrchestrator,
)
from libraries.ontology.bnop.rdf_jena.orchestrators.sparql_query_orchestrator import (
    SparqlQueryOrchestrator,
)
from libraries.ontology.bnop.rdf_jena.rdf_jena_configuration_factory import (
    RdfJenaConfigurations,
    create_rdf_jena_configuration,
)


class RdfJenaServiceFacade:
    """Main entry point for RDF-Jena operations."""

    def __init__(
        self,
        configuration: RdfJenaConfigurations | None = None,
    ) -> None:
        self.configuration = configuration or create_rdf_jena_configuration()
        self.database = JenaDatabases(self.configuration.jena_config)
        self.session = JenaSessions(self.database)
        loader_cfg = self.configuration.loader_config
        self.data_loader = JenaDataLoadOrchestrator(
            self.session,
            batch_size=loader_cfg.batch_size,
            streaming_threshold=loader_cfg.streaming_threshold_mb,
        )
        self.query_runner = SparqlQueryOrchestrator(self.session)

    def load(
        self,
        source: Iterable[object],
        converter: BaseRdfConverter,
        *,
        strategy: str | None = None,
    ) -> None:
        """Convert *source* with *converter* and load into Jena."""
        self.data_loader.load_data(source, converter, strategy=strategy)

    def query(
        self,
        query: str,
        *,
        endpoints: Iterable[JenaSessions] | None = None,
    ) -> QueryResults:
        """Execute *query* and return structured results."""
        return self.query_runner.execute(query, endpoints=endpoints)

    def query_formatted(
        self,
        query: str,
        *,
        format: str = "json",
        endpoints: Iterable[JenaSessions] | None = None,
    ) -> str:
        """Execute *query* and return formatted results."""
        return self.query_runner.execute_formatted(
            query,
            format=format,
            endpoints=endpoints,
        )

    def close(self) -> None:
        """Close session and underlying database resources."""
        self.session.close()
        self.database.close()

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc, tb) -> None:  # pragma: no cover - defensive
        self.close()



"""RDF graph container and namespace manager."""

from dataclasses import dataclass, field

from .triples import Triple


@dataclass
class RdfGraph:
    """Collection of RDF triples with namespace and metadata support."""

    triples: list[Triple] = field(default_factory=list)
    namespaces: dict[str, str] = field(default_factory=dict)
    metadata: dict[str, str] = field(default_factory=dict)

    def add_triple(self, triple: Triple) -> None:
        """Append a triple to the graph."""
        self.triples.append(triple)

    def add_namespace(self, prefix: str, uri: str) -> None:
        """Register a namespace prefix."""
        self.namespaces[prefix] = uri

    def get_namespace(self, prefix: str) -> str:
        """Return the URI for a namespace prefix."""
        return self.namespaces[prefix]

    def to_n_triples(self) -> str:
        """Serialize all triples to N-Triples format."""
        return "\n".join(triple.to_n_triple() for triple in self.triples)



"""RDF-Jena integration package."""

from . import (
    configurations,
    constants,
    converters,
    object_models,
    orchestrators,
    resources,
)

__all__ = [
    "configurations",
    "constants",
    "converters",
    "object_models",
    "orchestrators",
    "resources",
]

__version__ = "0.1.0"



from __future__ import annotations

from urllib.parse import quote, urljoin

from libraries.ontology.bnop.rdf_jena.constants.rdf_data_types import (
    ValueType,
    get_literal_uri,
    is_valid_uri,
)


def build_uri(namespace: str, identifier: str) -> str:
    """Return full URI for *identifier* in *namespace*."""
    if not namespace:
        raise ValueError("namespace required")
    if not identifier:
        raise ValueError("identifier required")
    base = namespace
    if not base.endswith(("/", "#")):
        base += "#"
    uri = urljoin(base, quote(identifier))
    if not is_valid_uri(uri):
        raise ValueError(f"invalid uri: {uri}")
    return f"<{uri}>"


def map_literal(value: ValueType) -> str:
    """Return N-Triples literal for *value*."""
    uri = get_literal_uri(value)
    return f'"{value}"^^<{uri}>'


def ensure_uri(uri: str) -> None:
    """Validate *uri* string."""
    if uri.startswith("<") and uri.endswith(">"):
        uri = uri[1:-1]
    if not is_valid_uri(uri):
        raise ValueError(f"invalid uri: {uri}")


def format_error(
    message: str,
    *,
    line: int | None = None,
    column: int | None = None,
) -> str:
    """Return error message with location info."""
    parts = [message]
    if line is not None:
        parts.append(f"line {line}")
    if column is not None:
        parts.append(f"column {column}")
    if len(parts) == 1:
        return message
    return f"{message} ({', '.join(parts[1:])})"


def report_progress(processed: int, total: int) -> str:
    """Return formatted progress string."""
    if total <= 0:
        return f"{processed}/?"
    percent = (processed / total) * 100
    return f"{processed}/{total} ({percent:.1f}%)"



"""Batch loader for RDF resources."""

from __future__ import annotations

from collections.abc import Iterable, Mapping

from libraries.ontology.bnop.rdf_jena.object_models.jena_sessions import (
    JenaSessions,
)


class NodeLoadError(Exception):
    """Raised when node loading fails."""


class RdfNodeLoader:
    """Load RDF nodes into a Jena session."""

    def __init__(
        self,
        session: JenaSessions,
        batch_size: int = 100,
    ) -> None:
        self.session = session
        self.batch_size = batch_size

    def load_nodes(
        self,
        nodes: Iterable[Mapping[str, Mapping[str, str]]],
    ) -> None:
        """Load nodes with batch processing."""
        node_list = list(nodes)
        for i in range(0, len(node_list), self.batch_size):
            batch = node_list[i : i + self.batch_size]
            self._load_batch(batch)

    def _load_batch(
        self,
        batch: list[Mapping[str, Mapping[str, str]]],
    ) -> None:
        self.session.begin()
        try:
            for node in batch:
                self._validate_node(node)
                uri = node["uri"]
                props = node.get("properties", {})
                for predicate, obj in props.items():
                    self.session.add_triple(
                        uri,
                        predicate,
                        obj,
                    )
            self.session.commit()
        except Exception as exc:
            self.session.rollback()
            raise NodeLoadError(
                "failed to load nodes",
            ) from exc

    def _validate_node(
        self,
        node: Mapping[str, Mapping[str, str]],
    ) -> None:
        uri = node.get("uri")
        if not isinstance(uri, str):
            msg = "node 'uri' must be a string"
            raise TypeError(msg)
        if not uri:
            msg = "node 'uri' required"
            raise ValueError(msg)
        props = node.get("properties", {})
        if not isinstance(props, Mapping):
            msg = "'properties' must be a mapping"
            raise TypeError(msg)
        for pred, obj in props.items():
            if not pred or not obj:
                msg = "invalid predicate or object"
                raise ValueError(msg)



from __future__ import annotations

from collections.abc import Callable, Iterable, Mapping

from libraries.ontology.bnop.rdf_jena.object_models.jena_sessions import JenaSessions
from libraries.ontology.bnop.rdf_jena.orchestrators.rdf_edge_loader import RdfEdgeLoader
from libraries.ontology.bnop.rdf_jena.orchestrators.rdf_node_loader import RdfNodeLoader


class BulkLoadError(Exception):
    """Raised when bulk loading fails."""


class BulkLoadOrchestrator:
    """Coordinate high-performance RDF node and edge loading."""

    def __init__(self, session: JenaSessions, batch_size: int = 100) -> None:
        self.session = session
        self.node_loader = RdfNodeLoader(session, batch_size)
        self.edge_loader = RdfEdgeLoader(session, batch_size)

    def load(
        self,
        nodes: Iterable[Mapping[str, Mapping[str, str]]] | None = None,
        edges: Iterable[Mapping[str, str]] | None = None,
        progress: Callable[[int, int], None] | None = None,
    ) -> None:
        """Load nodes and edges with progress tracking."""
        node_list = list(nodes) if nodes is not None else []
        edge_list = list(edges) if edges is not None else []
        total = len(node_list) + len(edge_list)
        loaded = 0

        try:
            with self.session:
                if node_list:
                    self.node_loader.load_nodes(node_list)
                    loaded += len(node_list)
                    if progress:
                        progress(loaded, total)
                if edge_list:
                    self.edge_loader.load_edges(edge_list)
                    loaded += len(edge_list)
                    if progress:
                        progress(loaded, total)
        except Exception as exc:  # pragma: no cover - defensive programming
            raise BulkLoadError("failed to bulk load data") from exc



"""Manage namespace prefixes and URIs."""

from __future__ import annotations

from dataclasses import dataclass, field

from libraries.ontology.bnop.rdf_jena.constants.standard_namespaces import (
    STANDARD_NAMESPACES,
    build_uri,
)


@dataclass
class NamespaceManager:
    """Register and resolve namespace prefixes."""

    namespaces: dict[str, str] = field(
        default_factory=lambda: dict(
            STANDARD_NAMESPACES,
        ),
    )

    def register(
        self,
        prefix: str,
        namespace: str,
    ) -> None:
        """Register ``prefix`` with ``namespace``.

        Raises
        ------
        ValueError
            If ``prefix`` already exists with a
            different ``namespace``.

        """
        if prefix in self.namespaces and self.namespaces[prefix] != namespace:
            raise ValueError(
                f"prefix {prefix!r} already registered",
            )
        self.namespaces[prefix] = namespace

    def get_namespace(
        self,
        prefix: str,
    ) -> str:
        """Return namespace for ``prefix``."""
        return self.namespaces[prefix]

    def resolve(
        self,
        prefixed_name: str,
    ) -> str:
        """Expand ``prefix:name`` to full URI."""
        prefix, _, name = prefixed_name.partition(
            ":",
        )
        if not name:
            raise ValueError(
                "missing ':' in prefixed name",
            )
        namespace = self.get_namespace(prefix)
        return build_uri(namespace, name)

    def shrink(
        self,
        uri: str,
    ) -> str:
        """Reduce full ``uri`` to ``prefix:name``.

        Selects the *longest* matching namespace to handle nested namespaces and
        ensure that ``shrink`` is the inverse of :meth:`resolve` for registered
        prefixes.
        """
        longest_prefix: str | None = None
        longest_namespace = ""
        for prefix, namespace in self.namespaces.items():
            if uri.startswith(namespace) and len(namespace) > len(longest_namespace):
                longest_prefix = prefix
                longest_namespace = namespace

        if longest_prefix is None:
            raise ValueError(
                "URI not in registered namespaces",
            )

        local = uri[len(longest_namespace) :]
        return f"{longest_prefix}:{local}"



from __future__ import annotations

import json
from collections.abc import Mapping
from datetime import date, datetime
from pathlib import Path

from libraries.ontology.bnop.rdf_jena.converters.base_rdf_converter import (
    BaseRdfConverter,
)
from libraries.ontology.bnop.rdf_jena.converters.helpers.conversion_utils import (
    build_uri,
    map_literal,
)
from libraries.ontology.bnop.rdf_jena.object_models.rdf_graphs import (
    RdfGraph,
)
from libraries.ontology.bnop.rdf_jena.object_models.triples import (
    Triple,
)


class JsonToRdfConverter(BaseRdfConverter):
    """Convert JSON objects to RDF triples."""

    def __init__(self, namespace: str) -> None:
        super().__init__()
        if not namespace:
            raise ValueError("namespace required")
        self._namespace = namespace
        self._blank_counter = 0

    def convert_file(self, path: str | Path) -> RdfGraph:
        """Convert JSON file at *path* to RDF."""
        with open(path, encoding="utf-8") as handle:
            data = json.load(handle)
        items = data if isinstance(data, list) else [data]
        return self.convert(items)

    def _convert_item(
        self,
        item: Mapping[str, object],
        graph: RdfGraph,
    ) -> None:
        """Add triples for JSON *item* to *graph*."""
        if not isinstance(item, Mapping):
            raise TypeError("item must be mapping")
        identifier = str(item.get("id", self.processed))
        subject = build_uri(self._namespace, identifier)
        self._handle_mapping(subject, item, graph)

    def _handle_mapping(
        self,
        subject: str,
        mapping: Mapping[str, object],
        graph: RdfGraph,
    ) -> None:
        for key, value in mapping.items():
            if key == "id" or value is None:
                continue
            predicate = build_uri(self._namespace, key)
            self._handle_value(subject, predicate, value, graph)

    def _handle_value(
        self,
        subject: str,
        predicate: str,
        value: object,
        graph: RdfGraph,
    ) -> None:
        if isinstance(value, Mapping):
            blank = self._new_blank_node()
            graph.add_triple(Triple(subject, predicate, blank))
            self._handle_mapping(blank, value, graph)
        elif isinstance(value, list):
            for item in value:
                self._handle_value(subject, predicate, item, graph)
        else:
            if not isinstance(value, (str, bool, int, float, date, datetime)):
                value = str(value)
            obj = map_literal(value)  # type: ignore[arg-type]
            graph.add_triple(Triple(subject, predicate, obj))

    def _new_blank_node(self) -> str:
        self._blank_counter += 1
        return f"_:b{self._blank_counter}"



from __future__ import annotations

"""Constants and utilities for common SPARQL query forms."""

import string

# Supported SPARQL query forms
SELECT = "SELECT"
CONSTRUCT = "CONSTRUCT"
ASK = "ASK"
DESCRIBE = "DESCRIBE"

QUERY_FORMS: set[str] = {SELECT, CONSTRUCT, ASK, DESCRIBE}

# Minimal query templates for each form
QUERY_TEMPLATES: dict[str, str] = {
    SELECT: "SELECT {select} WHERE {where}",
    CONSTRUCT: "CONSTRUCT {construct} WHERE {where}",
    ASK: "ASK WHERE {where}",
    DESCRIBE: "DESCRIBE {describe} WHERE {where}",
}


def detect_query_form(query: str) -> str | None:
    """Return the SPARQL form used by ``query`` if supported."""
    stripped = query.lstrip().upper()
    for form in QUERY_FORMS:
        if stripped.startswith(form):
            return form
    return None


def is_supported_query(query: str) -> bool:
    """Return ``True`` if ``query`` starts with a supported SPARQL form."""
    return detect_query_form(query) is not None


def bind_query_template(template: str, **params: str) -> str:
    """Fill a SPARQL query ``template`` with ``params``.

    Raises ``KeyError`` if required parameters are missing.
    """
    formatter = string.Formatter()
    required = {
        field_name for _, field_name, _, _ in formatter.parse(template) if field_name
    }
    missing = required - params.keys()
    if missing:
        raise KeyError(f"Missing parameters: {sorted(missing)}")
    return template.format(**params)



"""Factory for composing RDF-Jena configuration."""

from dataclasses import dataclass

from bclearer_core.configurations.b_clearer_configurations.b_clearer_configurations import (
    BClearerConfigurations,
)

from libraries.ontology.bnop.rdf_jena.configurations import (
    JenaConfigurations,
    JenaLoaderConfigurations,
    RdfConfigurations,
    SparqlConfigurations,
)


@dataclass
class RdfJenaConfigurations(BClearerConfigurations):
    """Combined configuration for RDF-Jena."""

    jena_config: JenaConfigurations
    rdf_config: RdfConfigurations
    sparql_config: SparqlConfigurations
    loader_config: JenaLoaderConfigurations


def create_rdf_jena_configuration(
    jena_config: JenaConfigurations | None = None,
    rdf_config: RdfConfigurations | None = None,
    sparql_config: SparqlConfigurations | None = None,
    loader_config: JenaLoaderConfigurations | None = None,
) -> RdfJenaConfigurations:
    """Create a complete RDF-Jena configuration."""
    jena_cfg = jena_config or JenaConfigurations.create_in_memory()
    rdf_cfg = rdf_config or RdfConfigurations()
    sparql_cfg = sparql_config or SparqlConfigurations()
    loader_cfg = loader_config or JenaLoaderConfigurations()

    return RdfJenaConfigurations(
        jena_config=jena_cfg,
        rdf_config=rdf_cfg,
        sparql_config=sparql_cfg,
        loader_config=loader_cfg,
    )



"""Helper classes for RDF-Jena orchestrators."""

from .namespace_manager import NamespaceManager
from .rdf_validator import RdfValidator
from .serializer import Serializer

__all__ = [
    "NamespaceManager",
    "RdfValidator",
    "Serializer",
]



"""Orchestrators coordinating RDF-Jena workflows."""

from . import helpers
from .bulk_load_orchestrator import (
    BulkLoadError,
    BulkLoadOrchestrator,
)
from .jena_data_load_orchestrator import (
    JenaDataLoadOrchestrator,
)
from .rdf_edge_loader import (
    EdgeLoadError,
    RdfEdgeLoader,
)
from .rdf_node_loader import (
    NodeLoadError,
    RdfNodeLoader,
)
from .sparql_query_orchestrator import (
    SparqlQueryError,
    SparqlQueryOrchestrator,
)
from .streaming_orchestrator import (
    StreamingError,
    StreamingOrchestrator,
)

__all__ = [
    "BulkLoadError",
    "BulkLoadOrchestrator",
    "EdgeLoadError",
    "JenaDataLoadOrchestrator",
    "NodeLoadError",
    "RdfEdgeLoader",
    "RdfNodeLoader",
    "SparqlQueryError",
    "SparqlQueryOrchestrator",
    "StreamingError",
    "StreamingOrchestrator",
    "helpers",
]



from __future__ import annotations

from dataclasses import dataclass, field
from typing import ClassVar

from bclearer_core.configurations.b_clearer_configurations.b_clearer_configurations import (
    BClearerConfigurations,
)

from libraries.ontology.bnop.rdf_jena.constants.serialization_formats import (
    SUPPORTED_FORMATS,
)
from libraries.ontology.bnop.rdf_jena.constants.standard_namespaces import (
    BORO_NS,
    STANDARD_NAMESPACES,
)


@dataclass
class RdfConfigurations(BClearerConfigurations):
    """Settings controlling RDF generation and namespaces."""

    default_namespace: str = BORO_NS
    generate_blank_nodes: bool = True
    preserve_boro_uuids: bool = True
    namespaces: dict[str, str] = field(
        default_factory=lambda: STANDARD_NAMESPACES.copy(),
    )
    supported_formats: list[str] = field(
        default_factory=lambda: list(SUPPORTED_FORMATS.keys()),
    )

    VALID_NAMESPACE_PREFIXES: ClassVar[tuple[str, ...]] = ("http://", "https://")
    _NS_SUFFIXES: ClassVar[tuple[str, ...]] = ("#", "/")

    def __post_init__(self) -> None:
        self.validate()

    def add_namespace(self, prefix: str, uri: str) -> None:
        """Add a custom namespace mapping after validation."""
        self._validate_namespace_uri(uri, f"namespace '{prefix}'")
        self.namespaces[prefix] = uri

    def validate(self) -> None:
        if not self.default_namespace:
            raise ValueError("default_namespace must not be empty")
        self._validate_namespace_uri(self.default_namespace, "default_namespace")
        if self.default_namespace not in self.namespaces.values():
            self.namespaces.setdefault("default", self.default_namespace)
        for prefix, uri in self.namespaces.items():
            self._validate_namespace_uri(uri, f"namespace '{prefix}'")
        for fmt in self.supported_formats:
            if fmt not in SUPPORTED_FORMATS:
                raise ValueError(f"unsupported format: {fmt}")

    def _validate_namespace_uri(self, uri: str, field_name: str) -> None:
        if not uri.startswith(self.VALID_NAMESPACE_PREFIXES):
            raise ValueError(
                f"{field_name} must start with 'http://' or 'https://'",
            )
        if not uri.endswith(self._NS_SUFFIXES):
            raise ValueError(
                f"{field_name} must end with '#' or '/'",
            )



from __future__ import annotations

from typing import Any, Self

from libraries.ontology.bnop.rdf_jena.configurations.jena_configurations import (
    JenaConfigurations,
)

from .jena_connections import JenaConnections


class JenaDatabases:
    """Handle dataset management and session
    creation for Jena stores.
    """

    def __init__(self, configuration: JenaConfigurations) -> None:
        self.configuration = configuration
        self.connections = JenaConnections(configuration)
        self._dataset: dict[str, str] | None = None

    def get_or_create_dataset(self) -> dict[str, str]:
        """Return existing dataset or create a
        new one.
        """
        if self._dataset is None:
            self._dataset = {
                "store_type": self.configuration.jena_store_type,
                "path": self.configuration.jena_store_path,
            }
        return self._dataset

    def create_session(self) -> dict[str, Any]:
        """Acquire a connection and bind it to the
        dataset.
        """
        dataset = self.get_or_create_dataset()
        connection = self.connections.acquire()
        return {"dataset": dataset, "connection": connection}

    def close_session(self, session: dict[str, Any]) -> None:
        """Release the session's connection."""
        connection = session.get("connection")
        if connection is not None:
            self.connections.release(connection)

    def close(self) -> None:
        """Close all database resources."""
        self.connections.close()

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc, tb) -> None:
        self.close()



"""Prefix mappings utilities for RDF output."""

from __future__ import annotations

from libraries.ontology.bnop.rdf_jena.constants.standard_namespaces import (
    STANDARD_NAMESPACES,
)

PREFIX_MAPPINGS: dict[str, str] = dict(STANDARD_NAMESPACES)


def register_prefix(prefix: str, namespace: str) -> None:
    """Register ``prefix`` for ``namespace``."""
    PREFIX_MAPPINGS[prefix] = namespace


def get_namespace(prefix: str) -> str | None:
    """Return namespace URI for ``prefix`` if registered."""
    return PREFIX_MAPPINGS.get(prefix)


def get_prefix(namespace: str) -> str | None:
    """Return prefix associated with ``namespace`` if registered."""
    for registered_prefix, registered_namespace in PREFIX_MAPPINGS.items():
        if registered_namespace == namespace:
            return registered_prefix
    return None



from __future__ import annotations

import csv
import io
import json
from dataclasses import dataclass, field
from typing import Any
from xml.etree.ElementTree import Element, SubElement, tostring


@dataclass
class QueryResults:
    """Structured SPARQL query results."""

    rows: list[dict[str, Any]] = field(default_factory=list)

    def add_row(self, row: dict[str, Any]) -> None:
        """Add a result row."""
        self.rows.append(row)

    def to_json(self) -> str:
        """Return results in JSON format."""
        return json.dumps({"results": self.rows})

    def to_csv(self) -> str:
        """Return results in CSV format."""
        if not self.rows:
            return ""
        output = io.StringIO()
        writer = csv.DictWriter(output, fieldnames=list(self.rows[0].keys()))
        writer.writeheader()
        writer.writerows(self.rows)
        return output.getvalue().strip()

    def to_xml(self) -> str:
        """Return results in XML format."""
        root = Element("results")
        for row in self.rows:
            result_el = SubElement(root, "result")
            for key, value in row.items():
                binding = SubElement(result_el, key)
                binding.text = str(value)
        return tostring(root, encoding="unicode")



"""Constants and utilities for RDF serialization formats."""

from __future__ import annotations

from os import path

# Supported serialization formats mapped to their MIME types
SUPPORTED_FORMATS: dict[str, str] = {
    "turtle": "text/turtle",
    "rdf/xml": "application/rdf+xml",
    "json-ld": "application/ld+json",
    "n-triples": "application/n-triples",
}

# File extensions mapped to serialization formats
EXTENSION_TO_FORMAT: dict[str, str] = {
    ".ttl": "turtle",
    ".rdf": "rdf/xml",
    ".xml": "rdf/xml",
    ".jsonld": "json-ld",
    ".json": "json-ld",
    ".nt": "n-triples",
}

# MIME types mapped to serialization formats
MIME_TYPE_TO_FORMAT: dict[str, str] = {
    mime: fmt for fmt, mime in SUPPORTED_FORMATS.items()
}


def get_mime_type(format_name: str) -> str:
    """Return the MIME type for a supported serialization format."""
    return SUPPORTED_FORMATS[format_name]


def detect_format_from_filename(filename: str) -> str | None:
    """Return the serialization format inferred from a file name.

    Returns ``None`` if the extension is not recognized.
    """
    _, ext = path.splitext(filename.lower())
    return EXTENSION_TO_FORMAT.get(ext)


def detect_format_from_mime_type(mime_type: str) -> str | None:
    """Return the serialization format for a MIME type.

    Returns ``None`` if the MIME type is not supported.
    """
    return MIME_TYPE_TO_FORMAT.get(mime_type.lower())



from __future__ import annotations

from dataclasses import dataclass, field
from typing import ClassVar

from bclearer_core.configurations.b_clearer_configurations.b_clearer_configurations import (
    BClearerConfigurations,
)


@dataclass
class SparqlConfigurations(BClearerConfigurations):
    """Settings controlling SPARQL query execution."""

    timeout_seconds: int = 30
    result_limit: int = 1000
    allow_federated_queries: bool = False
    service_endpoints: dict[str, str] = field(
        default_factory=dict,
    )

    VALID_URI_PREFIXES: ClassVar[tuple[str, ...]] = (
        "http://",
        "https://",
    )

    def __post_init__(self) -> None:
        self.validate()

    def add_endpoint(self, name: str, uri: str) -> None:
        """Add a federated SPARQL service endpoint."""
        self._validate_endpoint_uri(uri, f"service '{name}'")
        self.service_endpoints[name] = uri

    def validate(self) -> None:
        if self.timeout_seconds <= 0:
            raise ValueError("timeout_seconds must be positive")
        if self.result_limit <= 0:
            raise ValueError("result_limit must be positive")
        if not self.allow_federated_queries and self.service_endpoints:
            raise ValueError(
                "federated queries disabled but service_endpoints provided",
            )
        for name, uri in self.service_endpoints.items():
            self._validate_endpoint_uri(uri, f"service '{name}'")

    def _validate_endpoint_uri(self, uri: str, field_name: str) -> None:
        if not uri.startswith(self.VALID_URI_PREFIXES):
            raise ValueError(
                f"{field_name} must start with 'http://' or 'https://'",
            )



"""Common RDF and BORO namespaces with helpers."""

from __future__ import annotations

# Standard RDF/OWL namespaces
RDF_NS = "http://www.w3.org/1999/02/22-rdf-syntax-ns#"
RDFS_NS = "http://www.w3.org/2000/01/rdf-schema#"
OWL_NS = "http://www.w3.org/2002/07/owl#"
SKOS_NS = "http://www.w3.org/2004/02/skos/core#"
DCTERMS_NS = "http://purl.org/dc/terms/"

# BORO namespace for bclearer ontology objects
BORO_NS = "http://bclearer.org/ontology#"

# Mapping of common prefix to namespace URI
STANDARD_NAMESPACES: dict[str, str] = {
    "rdf": RDF_NS,
    "rdfs": RDFS_NS,
    "owl": OWL_NS,
    "skos": SKOS_NS,
    "dcterms": DCTERMS_NS,
    "boro": BORO_NS,
}


def build_uri(namespace: str, name: str) -> str:
    """Return a URI within ``namespace`` for ``name``."""
    return f"{namespace}{name}"


def build_boro_uri(name: str) -> str:
    """Return a URI within the BORO namespace for ``name``."""
    return build_uri(BORO_NS, name)



from __future__ import annotations

from pathlib import Path

from defusedxml import ElementTree

from libraries.ontology.bnop.rdf_jena.converters.base_rdf_converter import (
    BaseRdfConverter,
)
from libraries.ontology.bnop.rdf_jena.converters.helpers.conversion_utils import (
    build_uri,
    map_literal,
)
from libraries.ontology.bnop.rdf_jena.object_models.rdf_graphs import (
    RdfGraph,
)
from libraries.ontology.bnop.rdf_jena.object_models.triples import (
    Triple,
)


class XmlToRdfConverter(BaseRdfConverter):
    """Convert XML elements to RDF triples."""

    def __init__(self, namespace: str) -> None:
        super().__init__()
        if not namespace:
            raise ValueError("namespace required")
        self._namespace = namespace
        self._blank_counter = 0

    def convert_file(self, path: str | Path) -> RdfGraph:
        """Convert XML file at *path* to RDF."""
        namespaces: dict[str, str] = {}
        for _, elem in ElementTree.iterparse(path, events=("start-ns",)):
            prefix, uri = elem
            namespaces[prefix] = uri
        tree = ElementTree.parse(path)
        root = tree.getroot()
        graph = self.convert([root])
        for prefix, uri in namespaces.items():
            graph.add_namespace(prefix, uri)
        return graph

    def _convert_item(self, item: ElementTree.Element, graph: RdfGraph) -> None:  # type: ignore[override]
        """Add triples for XML *item* to *graph*."""
        subject = self._element_uri(item)
        for attr, value in item.attrib.items():
            if attr == "id":
                continue
            predicate = build_uri(self._namespace, self._local_name(attr))
            obj = map_literal(value)
            graph.add_triple(Triple(subject, predicate, obj))
        text = (item.text or "").strip()
        if text:
            predicate = build_uri(self._namespace, "value")
            obj = map_literal(text)
            graph.add_triple(Triple(subject, predicate, obj))
        for child in list(item):
            child_subject = self._element_uri(child)
            predicate = build_uri(self._namespace, self._local_name(child.tag))
            graph.add_triple(Triple(subject, predicate, child_subject))
            self._convert_item(child, graph)

    def _element_uri(self, element: ElementTree.Element) -> str:
        identifier = element.attrib.get("id")
        if identifier:
            return build_uri(self._namespace, identifier)
        self._blank_counter += 1
        return f"_:b{self._blank_counter}"

    @staticmethod
    def _local_name(tag: str) -> str:
        return tag.split("}")[-1]



from __future__ import annotations

from collections.abc import Iterable

from libraries.ontology.bnop.rdf_jena.object_models.jena_sessions import (
    JenaSessions,
)
from libraries.ontology.bnop.rdf_jena.object_models.query_results import (
    QueryResults,
)


class SparqlQueryError(Exception):
    """Raised when SPARQL query execution fails."""


class SparqlQueryOrchestrator:
    """Execute SPARQL queries and format results."""

    def __init__(self, session: JenaSessions) -> None:
        self.session = session

    def execute(
        self,
        query: str,
        *,
        endpoints: Iterable[JenaSessions] | None = None,
    ) -> QueryResults:
        """Run *query* against session and optional *endpoints*."""
        results = QueryResults()
        datasets = [self.session]
        if endpoints:
            datasets.extend(endpoints)
        for dataset in datasets:
            try:
                for subj, pred, obj in dataset.query(query):
                    row = {
                        "subject": subj,
                        "predicate": pred,
                        "object": obj,
                    }
                    if endpoints:
                        row["endpoint"] = (
                            "local" if dataset is self.session else "federated"
                        )
                    results.add_row(row)
            except Exception as exc:  # pragma: no cover - defensive
                raise SparqlQueryError("failed to execute query") from exc
        return results

    def execute_formatted(
        self,
        query: str,
        *,
        format: str = "json",
        endpoints: Iterable[JenaSessions] | None = None,
    ) -> str:
        """Run *query* and return results in selected *format*."""
        results = self.execute(query, endpoints=endpoints)
        if format == "json":
            return results.to_json()
        if format == "csv":
            return results.to_csv()
        if format == "xml":
            return results.to_xml()
        msg = "format must be 'json', 'csv', or 'xml'"
        raise ValueError(msg)



import pytest
from bnop.rdf_jena.constants.serialization_formats import (
    EXTENSION_TO_FORMAT,
    MIME_TYPE_TO_FORMAT,
    SUPPORTED_FORMATS,
    detect_format_from_filename,
    detect_format_from_mime_type,
    get_mime_type,
)


def test_supported_formats_keys():
    assert set(SUPPORTED_FORMATS) == {
        "turtle",
        "rdf/xml",
        "json-ld",
        "n-triples",
    }


@pytest.mark.parametrize(
    ("format_name", "mime_type"),
    [
        ("turtle", "text/turtle"),
        ("rdf/xml", "application/rdf+xml"),
        ("json-ld", "application/ld+json"),
        ("n-triples", "application/n-triples"),
    ],
)
def test_get_mime_type(format_name, mime_type):
    assert get_mime_type(format_name) == mime_type


@pytest.mark.parametrize(
    ("filename", "expected"),
    [
        ("example.ttl", "turtle"),
        ("example.rdf", "rdf/xml"),
        ("example.xml", "rdf/xml"),
        ("example.jsonld", "json-ld"),
        ("example.json", "json-ld"),
        ("example.nt", "n-triples"),
    ],
)
def test_detect_format_from_filename(filename, expected):
    assert detect_format_from_filename(filename) == expected


def test_detect_format_from_filename_unknown():
    assert detect_format_from_filename("example.txt") is None


@pytest.mark.parametrize(
    ("mime_type", "expected"),
    [
        ("text/turtle", "turtle"),
        ("application/rdf+xml", "rdf/xml"),
        ("application/ld+json", "json-ld"),
        ("application/n-triples", "n-triples"),
    ],
)
def test_detect_format_from_mime_type(mime_type, expected):
    assert detect_format_from_mime_type(mime_type) == expected


def test_detect_format_from_mime_type_unknown():
    assert detect_format_from_mime_type("text/plain") is None



"""Reusable SPARQL query templates."""

from libraries.ontology.bnop.rdf_jena.constants.sparql_query_forms import (
    ASK,
    CONSTRUCT,
    DESCRIBE,
    QUERY_TEMPLATES,
    SELECT,
    bind_query_template,
)


def build_select_query(select: str, where: str) -> str:
    """Return a SELECT query with ``select`` and ``where``."""
    template = QUERY_TEMPLATES[SELECT]
    return bind_query_template(template, select=select, where=where)


def build_construct_query(construct: str, where: str) -> str:
    """Return a CONSTRUCT query."""
    template = QUERY_TEMPLATES[CONSTRUCT]
    return bind_query_template(template, construct=construct, where=where)


def build_ask_query(where: str) -> str:
    """Return an ASK query."""
    template = QUERY_TEMPLATES[ASK]
    return bind_query_template(template, where=where)


def build_describe_query(describe: str, where: str) -> str:
    """Return a DESCRIBE query."""
    template = QUERY_TEMPLATES[DESCRIBE]
    return bind_query_template(template, describe=describe, where=where)



"""Converters turning various sources into RDF."""

from . import helpers
from .base_rdf_converter import BaseRdfConverter
from .csv_to_rdf_converter import CsvToRdfConverter
from .database_to_rdf_converter import (
    DatabaseToRdfConverter,
)
from .graph_to_rdf_converter import GraphToRdfConverter
from .json_to_rdf_converter import JsonToRdfConverter
from .xml_to_rdf_converter import XmlToRdfConverter

__all__ = [
    "BaseRdfConverter",
    "CsvToRdfConverter",
    "DatabaseToRdfConverter",
    "GraphToRdfConverter",
    "JsonToRdfConverter",
    "XmlToRdfConverter",
    "helpers",
]



from __future__ import annotations

import json
import tempfile
from collections.abc import Callable, Iterable, Mapping

from libraries.ontology.bnop.rdf_jena.object_models.jena_sessions import (
    JenaSessions,
)
from libraries.ontology.bnop.rdf_jena.orchestrators.rdf_edge_loader import (
    RdfEdgeLoader,
)
from libraries.ontology.bnop.rdf_jena.orchestrators.rdf_node_loader import (
    RdfNodeLoader,
)


class StreamingError(Exception):
    """Raised when streaming operations fail."""


class StreamingOrchestrator:
    """Stream large RDF datasets efficiently."""

    def __init__(
        self,
        session: JenaSessions,
        batch_size: int = 100,
        cache_size_mb: int = 10,
    ) -> None:
        self.session = session
        self.batch_size = batch_size
        self.cache_size = cache_size_mb * 1024 * 1024
        self.node_loader = RdfNodeLoader(session, batch_size)
        self.edge_loader = RdfEdgeLoader(session, batch_size)

    def stream(
        self,
        nodes: Iterable[Mapping[str, Mapping[str, str]]] | None = None,
        edges: Iterable[Mapping[str, str]] | None = None,
    ) -> None:
        """Stream nodes and edges from iterables."""
        try:
            with self.session:
                if nodes:
                    self._stream_iter(nodes, self.node_loader.load_nodes)
                if edges:
                    self._stream_iter(edges, self.edge_loader.load_edges)
        except Exception as exc:  # pragma: no cover - defensive
            raise StreamingError("failed to stream data") from exc

    def _stream_iter(
        self,
        iterable: Iterable[Mapping[str, object]],
        loader: Callable[[Iterable[Mapping[str, object]]], None],
    ) -> None:
        with tempfile.SpooledTemporaryFile(
            max_size=self.cache_size,
            mode="w+",
            encoding="utf-8",
        ) as spooled:
            count = 0
            for item in iterable:
                spooled.write(json.dumps(item) + "\n")
                count += 1
                if count >= self.batch_size:
                    self._flush(spooled, loader)
                    count = 0
            if count:
                self._flush(spooled, loader)

    def _flush(
        self,
        spooled: tempfile.SpooledTemporaryFile,
        loader: Callable[[Iterable[Mapping[str, object]]], None],
    ) -> None:
        spooled.seek(0)
        batch = [json.loads(line) for line in spooled]
        loader(batch)
        spooled.seek(0)
        spooled.truncate()



from __future__ import annotations

from collections.abc import Iterable, Mapping

from libraries.ontology.bnop.rdf_jena.converters.base_rdf_converter import (
    BaseRdfConverter,
)
from libraries.ontology.bnop.rdf_jena.object_models.jena_sessions import (
    JenaSessions,
)
from libraries.ontology.bnop.rdf_jena.object_models.rdf_graphs import RdfGraph
from libraries.ontology.bnop.rdf_jena.orchestrators.bulk_load_orchestrator import (
    BulkLoadOrchestrator,
)
from libraries.ontology.bnop.rdf_jena.orchestrators.streaming_orchestrator import (
    StreamingOrchestrator,
)


class JenaDataLoadOrchestrator:
    """Orchestrate conversion and loading of RDF data."""

    def __init__(
        self,
        session: JenaSessions,
        batch_size: int = 100,
        streaming_threshold: int = 1000,
    ) -> None:
        self.session = session
        self.bulk = BulkLoadOrchestrator(session, batch_size)
        self.streaming = StreamingOrchestrator(session, batch_size)
        self.streaming_threshold = streaming_threshold

    def load_data(
        self,
        source: Iterable[object],
        converter: BaseRdfConverter,
        *,
        strategy: str | None = None,
    ) -> None:
        """Convert *source* and load it with selected strategy."""
        graph = converter.convert(source)
        nodes, edges = self._graph_to_loadables(graph)
        if strategy is None:
            total = len(nodes) + len(edges)
            strategy = "streaming" if total > self.streaming_threshold else "bulk"
        if strategy == "streaming":
            self.streaming.stream(nodes=nodes, edges=edges)
        elif strategy == "bulk":
            self.bulk.load(nodes=nodes, edges=edges)
        else:  # pragma: no cover - defensive
            msg = "strategy must be 'bulk' or 'streaming'"
            raise ValueError(msg)

    def _graph_to_loadables(
        self,
        graph: RdfGraph,
    ) -> tuple[
        list[Mapping[str, Mapping[str, str]]],
        list[Mapping[str, str]],
    ]:
        nodes: dict[str, dict[str, object]] = {}
        edges: list[dict[str, str]] = []
        for triple in graph.triples:
            subj = triple.subject
            pred = triple.predicate
            obj = triple.object
            if obj.startswith('"'):
                node = nodes.setdefault(
                    subj,
                    {"uri": subj, "properties": {}},
                )
                props = node["properties"]
                props[pred] = obj
            else:
                edges.append(
                    {
                        "subject": subj,
                        "predicate": pred,
                        "object": obj,
                    },
                )
        return list(nodes.values()), edges



"""Configuration classes for RDF-Jena."""

from .jena_configurations import JenaConfigurations
from .jena_loader_configurations import (
    JenaLoaderConfigurations,
)
from .rdf_configurations import RdfConfigurations
from .sparql_configurations import SparqlConfigurations

__all__ = [
    "JenaConfigurations",
    "JenaLoaderConfigurations",
    "RdfConfigurations",
    "SparqlConfigurations",
]



"""Low-level connection management for Jena stores."""

from __future__ import annotations

import time
from queue import Queue
from typing import Any, Self

from libraries.ontology.bnop.rdf_jena.configurations.jena_configurations import (
    JenaConfigurations,
)


class JenaConnectionError(Exception):
    """Raised when a Jena connection cannot be established."""


class JenaConnections:
    """Manage pooled connections to a Jena store."""

    def __init__(
        self,
        configuration: JenaConfigurations,
        pool_size: int = 1,
        max_retries: int = 3,
        retry_delay: float = 0.1,
    ) -> None:
        self.configuration = configuration
        self.pool: Queue[Any] = Queue(maxsize=pool_size)
        self.max_retries = max_retries
        self.retry_delay = retry_delay
        for _ in range(pool_size):
            self.pool.put(self._create_connection_with_retry())

    def _create_connection(self) -> dict[str, str]:
        """Create a raw connection object.

        Placeholder implementation to avoid heavy dependencies.
        """
        return {"store_type": self.configuration.jena_store_type}

    def _close_connection(self, connection: dict[str, str]) -> None:
        """Close a connection if necessary."""

    def _health_check(self, connection: dict[str, str]) -> bool:
        """Verify that the connection is alive."""
        return True

    def _create_connection_with_retry(self) -> dict[str, str]:
        for attempt in range(1, self.max_retries + 1):
            connection = self._create_connection()
            if self._health_check(connection):
                return connection
            time.sleep(self.retry_delay * attempt)
        raise JenaConnectionError("unable to establish connection")

    def acquire(self) -> dict[str, str]:
        """Get a connection from the pool."""
        connection = self.pool.get()
        if not self._health_check(connection):
            connection = self._create_connection_with_retry()
        return connection

    def release(self, connection: dict[str, str]) -> None:
        """Return a connection to the pool."""
        self.pool.put(connection)

    def close(self) -> None:
        """Close all pooled connections."""
        while not self.pool.empty():
            connection = self.pool.get()
            self._close_connection(connection)

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type,
        exc_value,
        traceback,
    ) -> None:
        self.close()



"""RDF graph validation utilities."""

from __future__ import annotations

from dataclasses import dataclass

from rdflib import Graph
from rdflib.namespace import OWL, RDF

try:  # pragma: no cover
    from pyshacl import validate as sh_validate
except ImportError:  # pragma: no cover
    sh_validate = None  # type: ignore[assignment]


@dataclass
class RdfValidator:
    """Validate graphs via SHACL and OWL."""

    def validate_shacl(
        self,
        data_graph: Graph,
        shacl_graph: Graph,
    ) -> None:
        """Validate with a SHACL shapes graph.

        Raises
        ------
        ImportError
            If pyshacl is missing.
        ValueError
            If validation fails.

        """
        if sh_validate is None:
            raise ImportError(
                "pyshacl is required",
            )
        conforms, _, report = sh_validate(
            data_graph,
            shacl_graph=shacl_graph,
        )
        if not conforms:
            raise ValueError(report)

    def validate_owl(
        self,
        graph: Graph,
    ) -> None:
        """Check basic OWL consistency."""
        for c1, _, c2 in graph.triples(
            (None, OWL.disjointWith, None),
        ):
            for inst in graph.subjects(
                RDF.type,
                c1,
            ):
                if (inst, RDF.type, c2) in graph:
                    raise ValueError(
                        "ontology inconsistent",
                    )



from __future__ import annotations

from collections.abc import Iterable, Mapping

from libraries.ontology.bnop.rdf_jena.object_models.jena_sessions import (
    JenaSessions,
)


class EdgeLoadError(Exception):
    """Raised when edge loading fails."""


class RdfEdgeLoader:
    """Load RDF edges into a Jena session."""

    def __init__(
        self,
        session: JenaSessions,
        batch_size: int = 100,
    ) -> None:
        self.session = session
        self.batch_size = batch_size

    def load_edges(
        self,
        edges: Iterable[Mapping[str, str]],
    ) -> None:
        """Load edges with batch processing."""
        edge_list = list(edges)
        for i in range(0, len(edge_list), self.batch_size):
            batch = edge_list[i : i + self.batch_size]
            self._load_batch(batch)

    def _load_batch(
        self,
        batch: list[Mapping[str, str]],
    ) -> None:
        self.session.begin()
        try:
            existing = set(self.session.query(""))
            for edge in batch:
                self._validate_edge(edge)
                subj = edge["subject"]
                pred = edge["predicate"]
                obj = edge["object"]
                triple = (subj, pred, obj)
                if triple in existing:
                    continue
                self.session.add_triple(
                    subj,
                    pred,
                    obj,
                )
                existing.add(triple)
            self.session.commit()
        except Exception as exc:
            self.session.rollback()
            raise EdgeLoadError(
                "failed to load edges",
            ) from exc

    def _validate_edge(
        self,
        edge: Mapping[str, str],
    ) -> None:
        for key in ("subject", "predicate", "object"):
            value = edge.get(key)
            if not isinstance(value, str):
                msg = f"edge '{key}' must be a string"
                raise TypeError(msg)
            if not value:
                msg = f"edge '{key}' required"
                raise ValueError(msg)



from datetime import UTC, date, datetime, timezone

import pytest
from bnop.rdf_jena.constants.rdf_data_types import (
    RDF_LITERAL_URIS,
    get_literal_uri,
    is_valid_literal,
    is_valid_uri,
)


@pytest.mark.parametrize(
    "uri",
    [
        "http://example.org/one",
        "https://example.org/two",
    ],
)
def test_is_valid_uri_accepts(uri):
    assert is_valid_uri(uri)


@pytest.mark.parametrize(
    "uri",
    [
        "example.org",
        "http://exa mple.org",
    ],
)
def test_is_valid_uri_rejects(uri):
    assert not is_valid_uri(uri)


def test_is_valid_literal_accepts():
    assert is_valid_literal("a", "string")
    assert is_valid_literal(
        value=True,
        literal="boolean",
    )
    assert is_valid_literal(7, "integer")
    assert is_valid_literal(3.5, "float")
    assert is_valid_literal(date(2024, 1, 1), "date")
    assert is_valid_literal(
        datetime(2024, 1, 1, tzinfo=UTC),
        "dateTime",
    )


def test_is_valid_literal_rejects_bool_as_int():
    assert not is_valid_literal(
        value=True,
        literal="integer",
    )


def test_get_literal_uri_returns():
    assert get_literal_uri("a") == RDF_LITERAL_URIS["string"]
    assert get_literal_uri(value=True) == RDF_LITERAL_URIS["boolean"]
    assert get_literal_uri(7) == RDF_LITERAL_URIS["integer"]
    assert get_literal_uri(3.5) == RDF_LITERAL_URIS["float"]
    assert get_literal_uri(date(2024, 1, 1)) == RDF_LITERAL_URIS["date"]
    assert (
        get_literal_uri(
            datetime(2024, 1, 1, tzinfo=UTC),
        )
        == RDF_LITERAL_URIS["dateTime"]
    )


def test_get_literal_uri_rejects_unsupported():
    with pytest.raises(TypeError):
        get_literal_uri(object())



"""Executable documentation examples for the RDF-Jena service."""

from __future__ import annotations

from collections.abc import Iterable

import networkx as nx

from libraries.ontology.bnop.rdf_jena.configurations import (
    JenaConfigurations,
    JenaLoaderConfigurations,
    RdfConfigurations,
    SparqlConfigurations,
)
from libraries.ontology.bnop.rdf_jena.converters.csv_to_rdf_converter import (
    CsvToRdfConverter,
)
from libraries.ontology.bnop.rdf_jena.converters.graph_to_rdf_converter import (
    GraphToRdfConverter,
)
from libraries.ontology.bnop.rdf_jena.converters.json_to_rdf_converter import (
    JsonToRdfConverter,
)
from libraries.ontology.bnop.rdf_jena.rdf_jena_configuration_factory import (
    create_rdf_jena_configuration,
)
from libraries.ontology.bnop.rdf_jena.rdf_jena_service_facade import (
    RdfJenaServiceFacade,
)

_SELECT_ALL_QUERY = (
    "SELECT ?subject ?predicate ?object WHERE { ?subject ?predicate ?object }"
)
_PEOPLE_NS = "http://example.org/people#"
_ORDERS_NS = "http://example.org/orders#"
_GRAPH_NS = "http://example.org/graph#"


def _create_service(
    *,
    namespace: str = _PEOPLE_NS,
    loader_threshold: int = 100,
    sparql: SparqlConfigurations | None = None,
) -> RdfJenaServiceFacade:
    """Return a service facade with in-memory configuration."""
    rdf_config = RdfConfigurations(default_namespace=namespace)
    loader_config = JenaLoaderConfigurations(
        streaming_threshold_mb=loader_threshold,
    )
    configuration = create_rdf_jena_configuration(
        jena_config=JenaConfigurations.create_in_memory(),
        rdf_config=rdf_config,
        loader_config=loader_config,
        sparql_config=sparql,
    )
    return RdfJenaServiceFacade(configuration)


def _sample_people() -> list[dict[str, str]]:
    return [
        {"id": "alice", "name": "Alice", "role": "Engineer"},
        {"id": "bob", "name": "Bob", "role": "Scientist"},
    ]


def load_csv_people_example() -> list[dict[str, str]]:
    """Load tabular records and return the resulting triples."""
    converter = CsvToRdfConverter(_PEOPLE_NS)
    with _create_service() as service:
        service.load(_sample_people(), converter)
        results = service.query(_SELECT_ALL_QUERY)
    return _sort_rows(results.rows)


def query_people_as_csv_example() -> str:
    """Run a SPARQL query and return CSV formatted results."""
    converter = CsvToRdfConverter(_PEOPLE_NS)
    with _create_service() as service:
        service.load(_sample_people(), converter)
        return service.query_formatted(
            _SELECT_ALL_QUERY,
            format="csv",
        )


def json_blank_nodes_example() -> list[dict[str, str]]:
    """Convert nested JSON structures and stream the load."""
    document = [
        {
            "id": "order-1",
            "customer": {"name": "Alice"},
            "items": [
                {"sku": "A1", "quantity": 2},
                {"sku": "B2", "quantity": 1},
            ],
        },
    ]
    converter = JsonToRdfConverter(_ORDERS_NS)
    with _create_service(
        namespace=_ORDERS_NS,
        loader_threshold=0,
    ) as service:
        service.load(
            document,
            converter,
            strategy="streaming",
        )
        results = service.query(_SELECT_ALL_QUERY)
    return _sort_rows(results.rows)


def graph_to_rdf_example() -> list[str]:
    """Create RDF triples from a NetworkX graph."""
    graph = nx.Graph()
    graph.add_node("A", label="Node A")
    graph.add_node("B", label="Node B")
    graph.add_edge("A", "B", label="connectedTo")
    converter = GraphToRdfConverter(_GRAPH_NS)
    rdf_graph = converter.convert_graph(graph)
    triples = rdf_graph.to_n_triples().splitlines()
    return sorted(triples)


def federated_query_example() -> dict[str, object]:
    """Return configured endpoints and combined query rows."""
    sparql_cfg = SparqlConfigurations(
        allow_federated_queries=True,
    )
    sparql_cfg.add_endpoint(
        "analytics",
        "http://analytics.example/sparql",
    )
    converter = CsvToRdfConverter(_PEOPLE_NS)
    with _create_service(sparql=sparql_cfg) as local:
        local.load(_sample_people(), converter)
        with _create_service() as remote:
            remote_converter = CsvToRdfConverter(_PEOPLE_NS)
            remote.load(
                [{"id": "carol", "name": "Carol"}],
                remote_converter,
            )
            results = local.query(
                _SELECT_ALL_QUERY,
                endpoints=[remote.session],
            )
    return {
        "endpoints": dict(
            local.configuration.sparql_config.service_endpoints,
        ),
        "rows": _sort_rows(
            results.rows,
            include_endpoint=True,
        ),
    }


def _sort_rows(
    rows: Iterable[dict[str, str]],
    *,
    include_endpoint: bool = False,
) -> list[dict[str, str]]:
    """Return rows sorted for deterministic documentation output."""
    key_fields = [
        "subject",
        "predicate",
        "object",
    ]
    if include_endpoint:
        key_fields.insert(0, "endpoint")
    return sorted(
        rows,
        key=lambda row: tuple(row.get(field, "") for field in key_fields),
    )


__all__ = [
    "federated_query_example",
    "graph_to_rdf_example",
    "json_blank_nodes_example",
    "load_csv_people_example",
    "query_people_as_csv_example",
]



from __future__ import annotations

from collections.abc import Mapping
from typing import Any

import networkx as nx

from libraries.ontology.bnop.rdf_jena.converters.base_rdf_converter import (
    BaseRdfConverter,
)
from libraries.ontology.bnop.rdf_jena.converters.helpers.conversion_utils import (
    build_uri,
    map_literal,
)
from libraries.ontology.bnop.rdf_jena.object_models.rdf_graphs import (
    RdfGraph,
)
from libraries.ontology.bnop.rdf_jena.object_models.triples import (
    Triple,
)

try:  # pragma: no cover - optional dependency
    from libraries.interop_services.bclearer_interop_services.b_simple_graph_service.objects.b_simple_graphs import (
        BSimpleGraphs,
    )
except Exception:  # pragma: no cover

    class BSimpleGraphs:  # type: ignore[too-few-public-methods]
        """Fallback stand-in when b_simple_graphs is unavailable."""


class GraphToRdfConverter(BaseRdfConverter):
    """Convert graph structures to RDF triples."""

    def __init__(self, namespace: str) -> None:
        super().__init__()
        if not namespace:
            raise ValueError("namespace required")
        self._namespace = namespace

    def convert_graph(self, graph: nx.Graph | BSimpleGraphs) -> RdfGraph:
        """Convert *graph* to an RDF graph."""
        if graph is None:
            raise ValueError("graph cannot be None")

        nx_graph = graph.graph if isinstance(graph, BSimpleGraphs) else graph
        rdf_graph = self.convert(nx_graph.nodes(data=True))
        self._add_edges(nx_graph, rdf_graph)
        return rdf_graph

    def _convert_item(
        self,
        item: tuple[Any, Mapping[str, Any]],
        graph: RdfGraph,
    ) -> None:
        """Add triples for node *item* to *graph*."""
        node_id, attributes = item
        subject = build_uri(self._namespace, str(node_id))
        for key, value in attributes.items():
            if value is None:
                continue
            predicate = build_uri(self._namespace, key)
            obj = map_literal(value)
            graph.add_triple(Triple(subject, predicate, obj))

    def _add_edges(self, graph: nx.Graph, rdf_graph: RdfGraph) -> None:
        for source, target, data in graph.edges(data=True):
            predicate_name = data.get("label", "relatedTo")
            predicate = build_uri(self._namespace, str(predicate_name))
            subject = build_uri(self._namespace, str(source))
            obj = build_uri(self._namespace, str(target))
            rdf_graph.add_triple(Triple(subject, predicate, obj))



from dataclasses import dataclass
from typing import ClassVar

from bclearer_core.configurations.b_clearer_configurations.b_clearer_configurations import (
    BClearerConfigurations,
)


@dataclass
class JenaConfigurations(BClearerConfigurations):
    """Settings for connecting to Jena stores."""

    jena_store_type: str
    jena_store_path: str
    enable_reasoning: bool = False
    reasoner_type: str = "OWL_DL"

    VALID_STORE_TYPES: ClassVar[set[str]] = {"TDB2", "IN_MEMORY"}

    def __post_init__(self) -> None:
        self.validate()

    def validate(self) -> None:
        if self.jena_store_type not in self.VALID_STORE_TYPES:
            raise ValueError(
                "jena_store_type must be 'TDB2' or 'IN_MEMORY'",
            )
        if self.jena_store_type == "TDB2" and not self.jena_store_path:
            raise ValueError("jena_store_path required for TDB2 store")
        if self.jena_store_type == "IN_MEMORY":
            self.jena_store_path = ""

    @classmethod
    def create_tdb2(
        cls,
        store_path: str,
        enable_reasoning: bool = False,
        reasoner_type: str = "OWL_DL",
    ) -> "JenaConfigurations":
        return cls(
            jena_store_type="TDB2",
            jena_store_path=store_path,
            enable_reasoning=enable_reasoning,
            reasoner_type=reasoner_type,
        )

    @classmethod
    def create_in_memory(
        cls,
        enable_reasoning: bool = False,
        reasoner_type: str = "OWL_DL",
    ) -> "JenaConfigurations":
        return cls(
            jena_store_type="IN_MEMORY",
            jena_store_path="",
            enable_reasoning=enable_reasoning,
            reasoner_type=reasoner_type,
        )



from __future__ import annotations

from collections.abc import Iterable
from typing import Any, Self

from .jena_databases import JenaDatabases


class TransactionError(Exception):
    """Raised when transaction operations fail."""


class JenaSessions:
    """Provide transactional operations for a Jena dataset."""

    def __init__(self, databases: JenaDatabases) -> None:
        self.databases = databases
        self._session: dict[str, Any] | None = None
        self._in_transaction = False
        self._snapshot: set[tuple[str, str, str]] | None = None

    def open(self) -> None:
        """Open a session if none is active."""
        if self._session is None:
            self._session = self.databases.create_session()
            dataset = self._session["dataset"]
            dataset.setdefault("data", set())

    def close(self) -> None:
        """Close the current session."""
        if self._session is not None:
            self.databases.close_session(self._session)
            self._session = None
            self._in_transaction = False
            self._snapshot = None

    def __enter__(self) -> Self:
        self.open()
        return self

    def __exit__(self, exc_type, exc, tb) -> None:
        self.close()

    def begin(self) -> None:
        """Begin a transaction."""
        if self._session is None:
            self.open()
        dataset = self._session["dataset"]
        self._snapshot = set(dataset["data"])
        self._in_transaction = True

    def commit(self) -> None:
        """Commit changes made during the transaction."""
        if not self._in_transaction:
            raise TransactionError("no active transaction")
        self._in_transaction = False
        self._snapshot = None

    def rollback(self) -> None:
        """Revert changes made during the transaction."""
        if not self._in_transaction:
            raise TransactionError("no active transaction")
        dataset = self._session["dataset"]
        dataset["data"] = self._snapshot or set()
        self._in_transaction = False
        self._snapshot = None

    def add_triple(self, subj: str, pred: str, obj: str) -> None:
        """Insert a triple into the dataset."""
        if self._session is None:
            self.open()
        dataset = self._session["dataset"]
        dataset["data"].add((subj, pred, obj))

    def delete_triple(self, subj: str, pred: str, obj: str) -> None:
        """Remove a triple from the dataset."""
        if self._session is None:
            self.open()
        dataset = self._session["dataset"]
        dataset["data"].discard((subj, pred, obj))

    def update_triple(
        self,
        old: tuple[str, str, str],
        new: tuple[str, str, str],
    ) -> None:
        """Replace one triple with another."""
        self.delete_triple(*old)
        self.add_triple(*new)

    def query(self, _sparql: str) -> Iterable[tuple[str, str, str]]:
        """Execute a SPARQL query and return results."""
        if self._session is None:
            self.open()
        dataset = self._session["dataset"]
        return set(dataset["data"])
